/*
 *  linkedlist.c
 *
 *  Description:
 *  	Hand-over-hand locking linked list implementation
 */

#include "linkedlist.h"

int list_contains(llist_t* the_list, val_t val)
{
	int key = hashFunction(val);
	node_t* predEntry;
	node_t* currEntry;
	
	predEntry = the_list->root;
	
	currEntry = predEntry->_next;
	if (currEntry == NULL)
		return 0;

	int currEntryKey = hashFunction(currEntry->_value);

	while (currEntryKey <= key) {
		if (val == currEntry->_value) {
			return 1;
		}
		predEntry = currEntry;
		currEntry = currEntry->_next;
		if (currEntry == NULL)
			break;
		currEntryKey = hashFunction(currEntry->_value);
	}
	return 0;
}


node_t* new_node(val_t val, node_t *next)
{
	node_t* node = (node_t*)malloc(sizeof(node_t));
	node->_value = val;
	node->_key = hashFunction(val);
	node->_next = next;
	lock_init(&node->_lock);
	return node;
}

llist_t* list_new()
{
  llist_t* the_list = (llist_t*)malloc(sizeof(llist_t));
  the_list->root = new_node(-1, NULL);
  return the_list;
}

void list_delete(llist_t *the_list)
{
}

int list_size(llist_t* the_list)
{
	node_t* curEntry = the_list->root;

	int size = 0;
	while (curEntry->_next != NULL) {
		size++;
		curEntry = curEntry->_next;
	}
	return size;
}

int list_add(llist_t *the_list, val_t val)
{
	int key = hashFunction(val);
	node_t * predEntry;
	node_t * currEntry;
	
	// get root node (the one that is empty)
	predEntry = the_list->root;
	lock_lock(&predEntry->_lock);

	// and start scanning through the list to find the correct position to insert the new node.
	currEntry = predEntry->_next;
	if (currEntry != NULL) {
		lock_lock(&currEntry->_lock);
		int currEntryKey = hashFunction(currEntry->_value);

		while (currEntryKey <= key) {
			lock_unlock(&predEntry->_lock);
			predEntry = currEntry;
			currEntry = currEntry->_next;
			if (currEntry == NULL)
				break;
			lock_lock(&currEntry->_lock);
			currEntryKey = hashFunction(currEntry->_value);
		}
	}

	// create new node...
	predEntry->_next = new_node(val, currEntry);
	
	// release locks
	if (currEntry != NULL) {
		lock_unlock(&currEntry->_lock);
	}
	lock_unlock(&predEntry->_lock);
	return 1;
}

int list_remove(llist_t *the_list, val_t val)
{
	int key = hashFunction(val);
	node_t* predEntry;
	node_t* currEntry;
	
	predEntry = the_list->root;
	lock_lock(&predEntry->_lock);

	currEntry = predEntry->_next;

	if (currEntry == NULL) {
		lock_unlock(&predEntry->_lock);
		return 0;
	}

	lock_lock(&currEntry->_lock);
	int currEntryKey = hashFunction(currEntry->_value);

	while (currEntryKey <= key) {
		if (val == currEntry->_value) {
			predEntry->_next = currEntry->_next;
			lock_unlock(&currEntry->_lock);
			lock_unlock(&predEntry->_lock);
			return 1;
		}
		lock_unlock(&predEntry->_lock);
		predEntry = currEntry;
		currEntry = currEntry->_next;
		if (currEntry == NULL) {
			break;
		}
		lock_lock(&currEntry->_lock);
		currEntryKey = hashFunction(currEntry->_value);
	}
	if (currEntry != NULL) {
		lock_unlock(&currEntry->_lock);
	}
	lock_unlock(&predEntry->_lock);
	return 0;
}

int hashFunction(val_t value) {
	return value;
}